import datetime
import logging
from typing import Dict, List

import chevron
import pendulum
import sqlalchemy as sa
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship

from workspaces.system.schema import AbstractFile, AbstractRequest
from workspaces.workflow.schema_interfaces import (
    WorkflowIF,
    WorkflowRequestFileIF,
    WorkflowRequestIF,
    WorkflowTemplateIF,
)

logger = logging.getLogger(__name__)

Base = declarative_base()


class Workflow(Base, WorkflowIF):
    """
    A Workflow is a suite of tasks defined by templates that must be
    executed together with some user-submitted files.
    """

    __tablename__ = "workflows"
    workflow_name = sa.Column("workflow_name", sa.String, primary_key=True)
    uses_casa = sa.Column("uses_casa", sa.Boolean, default=False, server_default="false", nullable=False)
    templates = relationship("WorkflowTemplate", backref="workflows", lazy="dynamic")
    requests = relationship("WorkflowRequest", back_populates="workflow")

    def __repr__(self):
        return f"<Workflow workflow_name={self.workflow_name}>"

    def render_templates(self, argument: Dict) -> List[AbstractFile]:
        """
        Render the templates associated with this workflow

        :param argument: the workflow argument JSON
        :param files:    the files to be processed
        :return:         a list of rendered template files
        """
        rendered_files = []

        """
        
        Original:
          - All the user's input files are collected at submission time
          - This method combines user input files with rendered template results
          - They wind up mixed together in the list on this request
        
        New:
          - User's input files are collected prior to submission
          - They already exist in the list of files on this request
          - At submission time, we only need to render the templates and append them to the request        
        """
        for template in self.templates:
            rendered_files.append(template.render(argument))

        return rendered_files

    # Pyramid support method: must accept a "request" argument that is unused by us
    def __json__(self, request=None) -> dict:
        return {
            "type": self.__class__.__name__,
            "workflow_name": self.workflow_name,
            "uses_casa": self.uses_casa,
            "templates": list(self.templates),  # <- the list() here is necessitated by lazy loading
        }

    @classmethod
    def from_json(cls, json: dict) -> any:
        return Workflow(json["workflow_name"], json["uses_casa"], json["templates"])


class WorkflowTemplate(Base, WorkflowTemplateIF):
    """
    Workflow Templates are Mustache-formatted files associated with a given Workflow
    """

    __tablename__ = "workflow_templates"
    filename = sa.Column("filename", sa.String, primary_key=True)
    content = sa.Column("content", sa.LargeBinary, nullable=False)
    workflow_name = sa.Column(
        "workflow_name",
        sa.String,
        sa.ForeignKey("workflows.workflow_name"),
        primary_key=True,
    )

    def render(self, argument: Dict) -> AbstractFile:
        """
        Render a given Mustache-formatted Workflow Template with the specified arguments.

        :param argument:    argument(s) needed for template render
        :return:            an AbstractFile rendered from mustache template
        """
        contents = chevron.render(self.content.decode("ascii"), argument)
        logger.info(
            "rendering %s with argument %s to %s",
            self.content.decode("ascii"),
            argument,
            contents,
        )
        return AbstractFile(self.filename, contents.encode("ascii"))

    @property
    def template_file(self) -> AbstractFile:
        return AbstractFile(self.filename, self.content)

    @template_file.setter
    def template_file(self, file: AbstractFile):
        self.filename, self.content = file.filename, file.content

    # Pyramid support method: must accept a "request" argument that is unused by us
    def __json__(self, request=None) -> dict:
        # Defer to the actual template file's contents for JSON conversion
        return {
            "filename": self.filename,
            "content": self.content.decode("utf8"),
        }

    @classmethod
    def from_json(cls, json: dict) -> any:
        return WorkflowTemplate(json["filename"], json["content"])

    def __repr__(self):
        return f"<WorkflowTemplate filename={self.filename} for workflow={self.workflow_name}>"


class WorkflowRequest(Base, WorkflowRequestIF):
    """
    Workflow Requests invoke a particular workflow with a particular argument and some user-supplied files.
    """

    __tablename__ = "workflow_requests"
    workflow_request_id = sa.Column("workflow_request_id", sa.Integer, primary_key=True, autoincrement=True)
    htcondor_job_id = sa.Column("htcondor_job_id", sa.Integer, default=None, server_default="null")
    workflow_name = sa.Column("workflow_name", sa.String, sa.ForeignKey("workflows.workflow_name"))
    argument = sa.Column("argument", sa.JSON)
    state = sa.Column("state", sa.String)
    results_dir = sa.Column("results_dir", sa.String)
    files = relationship("WorkflowRequestFile", backref="request")
    workflow = relationship("Workflow", back_populates="requests")
    created_at = sa.Column(
        "created_at",
        sa.DateTime(timezone=True),
        server_default=sa.func.now(),
        nullable=False,
    )
    updated_at = sa.Column(
        "updated_at",
        sa.DateTime(timezone=True),
        server_default=sa.func.now(),
        server_onupdate=sa.func.now(),
        nullable=False,
    )

    @property
    def created_at_ago(self) -> str:
        return pendulum.instance(self.created_at).diff_for_humans()

    @property
    def updated_at_ago(self) -> str:
        return pendulum.instance(self.updated_at).diff_for_humans()

    @property
    def request(self) -> AbstractRequest:
        return AbstractRequest(self.workflow_name, self.argument)

    @request.setter
    def create_request(self, request: AbstractRequest):
        self.workflow_name, self.argument = request.name, request.argument

    def update_status(self, status: str):
        self.state = status

    # TODO: create following fields in table
    def set_start_time(self, time: str):
        self.start_time = time

    def set_end_time(self, time: str):
        self.end_time = time

    def __getitem__(self, item):
        return self.files

    def __repr__(self):
        return f"<WorkflowRequest workflow_request_id={self.workflow_request_id}>"

    # Pyramid support method: must accept a "request" argument that is unused by us
    def __json__(self, request=None) -> dict:
        return dict(
            type=self.__class__.__name__,
            workflow_request_id=self.workflow_request_id,
            workflow_name=self.workflow_name,
            argument=self.argument,
            state=self.state,
            results_dir=self.results_dir,
            files=self.files,
            created_at=self.created_at.isoformat(),
            updated_at=self.updated_at.isoformat(),
        )

    @classmethod
    def from_json(cls, json: dict) -> any:
        # clean up the dictionary a bit
        copy = dict(json)

        # type will confuse the constructor
        del copy["type"]

        # fix the types of the datetimes
        copy["created_at"] = datetime.datetime.fromisoformat(copy["created_at"])
        copy["updated_at"] = datetime.datetime.fromisoformat(copy["updated_at"])

        # now return an instance
        return WorkflowRequest(**copy)


class WorkflowRequestFile(Base, WorkflowRequestFileIF):
    """
    A Workflow Request File is a file supplied by the user and attached to the request they have submitted.
    """

    def abstract_file(self):
        return AbstractFile(self.filename, self.content)

    __tablename__ = "workflow_request_files"
    workflow_request_id = sa.Column(
        "workflow_request_id",
        sa.Integer,
        sa.ForeignKey("workflow_requests.workflow_request_id"),
        primary_key=True,
    )
    filename = sa.Column("filename", sa.String, primary_key=True)
    content = sa.Column("content", sa.LargeBinary, nullable=False)

    @property
    def file(self) -> AbstractFile:
        return AbstractFile(self.filename, self.content)

    @file.setter
    def file(self, file: AbstractFile):
        self.filename, self.content = file.filename, file.content

    def __repr__(self):
        return f"<WorkflowRequestFile filename={self.filename}>"

    # Pyramid support method: must accept a "request" argument that is unused by us
    def __json__(self, request=None) -> dict:
        return {
            "type": self.__class__.__name__,
            "workflow_request_id": self.workflow_request_id,
            "filename": self.filename,
            "content": self.content.hex(),
        }

    @classmethod
    def from_json(cls, json: dict) -> any:
        return WorkflowRequestFile(json["filename"], bytes.fromhex(json["content"]))
